/*-
 * Copyright (c) 2006. Ivan Voras <ivoras@gmail.com>
 * Released under the Artistic License
 */

/*
 * sqlcached diagnostic/debugging/testing client
 * $Id: client.c,v 1.2 2006/04/30 22:31:48 ivoras Exp $
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <readline/readline.h>
#include <readline/history.h>

#include "common.h"
#include "sqlcached_client.h"

static char* unix_socket_name = DEFAULT_UNIX_SOCKET;

int argc;
char **argv;

struct sc_client_conn sc;
static char* whitespace = " \t";


/** Shows help message */
static void show_help() {
    printf("usage: %s [-h] | [-u unix_socket_name]\n", argv[0]);
    printf("        -h :                    Show this help message\n");
    printf("        -u unix_socket_name :   Set unix socket name of sqlcached (default %s)", DEFAULT_UNIX_SOCKET);
}


/** Parses command line arguments and validates state */
static void parse_args() {
    int ch;
    struct stat st;
    while ((ch = getopt(argc, argv, "hu:")) != -1) {
        switch (ch) {
        case 'h':
            show_help();
            break;
        case 'u':
            unix_socket_name = strdup(optarg);
            if (stat(unix_socket_name, &st) < 0) {
                fprintf(stderr, "Cannot access %s: %s\n", unix_socket_name, strerror(errno));
                exit(1);
            }
            break;
        }
    }
    if (stat(unix_socket_name, &st) < 0 || access(unix_socket_name, R_OK) < 0) {
        fprintf(stderr, "Cannot access %s: %s\n", unix_socket_name, strerror(errno));
        exit(1);
    }
}


/** Opens connection to the server */
static void open_connection() {
    int res;
    char *msg;

    if ((res = sc_connect_unix(unix_socket_name, &sc, &msg)) != SC_OK) {
        fprintf(stderr, "%s\n", msg);
        free(msg);
        exit(1);
    }

    printf("Diagnostic/debugging/testing client for %s\n", SC_VER_SIG);
    printf("Copyright (c) 2006. Ivan Voras <ivoras@gmail.com>\n");
}


void cmd_ver() {
    char sig[20];
    if (sc_get_ver(&sc, sig) != SC_OK) {
        fprintf(stderr, "Error getting version\n");
        exit(1);
    }
    printf("Server version signature: %s\n", sig);
}


void cmd_help() {
    printf("Lines not beginning with \".\" are treated as server SQL commands\n");
    printf("and will be passed as such to sqlcached. Lines beginning with \".\"\n");
    printf("should contain one of these commands:\n");
    printf("   .ver     Show server version signature\n");
    printf("   .help    Show this help\n");
    printf("   .quit    Quit this client application\n");
}


void cmd_quit() {
    printf("Quitting\n");
    exit(0);
}


void do_sql(char *sql) {
    int res, i, j;
    char *msg;
    struct sc_result_set rs;

    if ((res = sc_exec_query(&sc, sql, &rs, &msg)) != SC_OK) {
        fprintf(stderr, "error: %s\n", msg);
        return;
    }

    if (rs.n_rows > 0) {
        int ccols = 0;
        
        for (i = 0; i < rs.n_cols; i++)
            ccols += printf("|%s", rs.names[i]);
        printf("\n");

        for (i = 0; i < ccols; i++)
            printf("-");
        printf("\n");
        
        for (i = 0; i < rs.n_rows; i++) {
            for (j = 0; j < rs.n_cols; j++)
                printf("|%s", rs.rows[i][j]);
            printf("\n");
        }
    }

    printf("%d row(s).\n", rs.n_rows);
    sc_free_result(&sc, &rs);
}


/** Client loop */
void client_loop() {
    int done = 0;
    char *line;
    
    while (!done) {
        line = readline("sqlcached> ");
        if (strlen(line) != 0) {
            if (line[0] != '.') { /* treat it as sql */
                do_sql(line);
            } else { /* special/internal command */
                char *cmd, *param;
                cmd = strtok(line, whitespace);
                if (strcmp(cmd, ".ver") == 0)
                    cmd_ver();
                else if (strcmp(cmd, ".help") == 0)
                    cmd_help();
                else if (strcmp(cmd, ".quit") == 0)
                    cmd_quit();
                else {
                    param = strtok(NULL, whitespace);
                    printf("Huh?\n");
                }
            }
        }
        free(line);
    }

}


/** main() */
int main(int _argc, char **_argv) {
    argc = _argc;
    argv = _argv;
    parse_args();
    open_connection();
    client_loop();
    return 0;
}

