/*-
 * Copyright (c) 2006. Ivan Voras <ivoras@gmail.com>
 * Released under the Artistic License
 */

/*
 * sqlcached server
 * $Id: server.c,v 1.8 2006/10/12 13:33:24 ivoras Exp $
 */

#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <time.h>
#include <netinet/in.h>
#include <paths.h>
#include <netdb.h>
#include <errno.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <poll.h>
#include <syslog.h>
#include <signal.h>

#include <sqlite.h>

#define SQLCACHED_SERVER
#include "common.h"

#define READ_SIZE 4096


int argc;
char **argv;

unsigned int tcp_port = 0;   /* By default, don't listen on TCP */
char *unix_socket_name = DEFAULT_UNIX_SOCKET;
char *schema_filename = NULL;
char *expiry_filename = NULL;
int debug_mode = 0;

sqlite *sqlite_db = NULL;
int sock_unix = -1, sock_tcp = -1;
unsigned int num_clients = 0;
struct sc_conn *conn = NULL;    /* array */
unsigned int conn_size = 0;
struct sc_expiry *expiry = NULL;    /* array */
unsigned int expiry_size = 0;
unsigned int expiry_interval = 100;  /* check expiry data every X ops */
unsigned int op_count = 0;


/** Shows help message for command line usage */
static void show_help() {
    printf("usage: %s [-h] | [-s schema_file] [-p tcp_port] [-u unix_socket] [-d]\n", argv[0]);
    printf("                    [-e expiry_file]\n\n");
    printf("    -h              : Show this message\n");
    printf("    -s schema_file  : Load specified SQL schema file on startup\n");
    printf("    -p tcp_port     : Turn on listening on a TCP port (default is no TCP)\n");
    printf("    -u unix_socket  : Listen on this unix socket (default is %s)\n", DEFAULT_UNIX_SOCKET);
    printf("    -d              : Debug mode (stay in foreground)\n");
    printf("    -e expiry_file  : Load table data expiry descriptions from this file\n");
}


/** Parses command line arguments */
static void parse_args() {
    int ch;
    while ((ch = getopt(argc, argv, "hdp:u:s:e:")) != -1) {
        switch (ch) {
        case 'h':
            show_help();
            exit(0);
            break;
        case 'd':
            debug_mode++;
            break;
        case 'p':
            tcp_port = atoi(optarg);
            if (tcp_port == 0) {
                fprintf(stderr, "Invalid TCP port: %s\n", optarg);
                exit(1);
            }
            break;
        case 'u':
            unix_socket_name = strdup(optarg);
            break;
        case 's':
            schema_filename = strdup(optarg);
            if (access(schema_filename, R_OK) != 0) {
                fprintf(stderr, "Error accessing schema file %s: %s\n", schema_filename, strerror(errno));
                exit(1);
            }
            break;
        case 'e':
            expiry_filename = strdup(optarg);
            if (access(expiry_filename, R_OK) != 0) {
                fprintf(stderr, "Error accessing expiry file %s: %s\n", expiry_filename, strerror(errno));
                exit(1);
            }
            break;
        default:
            fprintf(stderr, "Unknown argument: %c\n", ch);
            show_help();
            exit(1);
        }
    }
}


#define STR_BUF_SIZE 256
/** Logging & error handling function */
void sc_log(int prio, char *msg, ...) {
    va_list ap;
    char *buf;

    buf = malloc(STR_BUF_SIZE);
    va_start(ap, msg);
    vsnprintf(buf, STR_BUF_SIZE-1, msg, ap);
    va_end(ap);

    if (debug_mode)
        fprintf(stderr, "%s\n", buf);
    else
        syslog(prio, "%s", buf);

    free(buf);
    if (prio == LOG_CRIT)
        exit(1);
}


/** Initialise logging */
void sc_log_init() {
    if (!debug_mode)
        openlog("sqlcached", LOG_PID, LOG_DAEMON);
}


/** Open/create the database, create initial schema */
static void start_database() {
    int res;
    int fd;
    struct stat sb;
    char *buf;
    char *errmsg;
    
    if ((sqlite_db = sqlite_open(":memory:", 0666, &errmsg)) == NULL) {
        sc_log(LOG_ERR, "Cannot initialise database: %s", errmsg);
        sqlite_close(sqlite_db);
        exit(1);
    }
   
    if (schema_filename != NULL) {
        if ((res = stat(schema_filename, &sb)) != 0 || access(schema_filename, R_OK) != 0)
            sc_log(LOG_CRIT, "Cannot stat %s: %s", schema_filename, strerror(errno));
        
        if ((fd = open(schema_filename, O_RDONLY)) < 0)
            sc_log(LOG_CRIT, "Cannot open schema file: %s", strerror(errno));

        buf = malloc(sb.st_size);
        if (buf == NULL)
            sc_log(LOG_CRIT, "Cannot malloc(%d): %s", (int)(sb.st_size), strerror(errno));

        if ((res = read(fd, buf, sb.st_size)) != sb.st_size)
            sc_log(LOG_CRIT, "Cannot read schema data: %s", strerror(errno));

        close(fd);
    } else {
        buf = 
            "CREATE TABLE cache ("
                "key    varchar not null,"
                "value  varchar not null,"
                "time   integer not null,"
                "primary key(key)"
            ");\n"
            "CREATE INDEX cache_time ON cache(time);"
            ;
    }

    if ((res = sqlite_exec(sqlite_db, buf, NULL, NULL, &errmsg)) != SQLITE_OK) {
        sc_log(LOG_CRIT, "Error executing initial schema: %s", errmsg);
        free(errmsg);
        exit(1);
    }

    if (schema_filename == NULL && expiry_filename == NULL) {
        expiry_size = 1;
        expiry = malloc(sizeof *expiry);
        strcpy(expiry[0].table_name, "cache");
        expiry[0].limit_ops = 10000;
        expiry[0].limit_time = 1800;
        expiry[0].next_ops = expiry[0].limit_ops;
        expiry[0].last_time = time(NULL);
        expiry[0].next_time = expiry[0].last_time + expiry[0].limit_time;
        expiry[0].limit_records = 0;
        expiry[0].last_rowid = 1;
    } else if (expiry_filename != NULL) {
        FILE *f;
        char table_name[104];
        unsigned int interval, ops, recs, line_no = 0;

        f = fopen(expiry_filename, "r");
        res = fscanf(f, "%u\n", &expiry_interval);
        if (res != 1)
            sc_log(LOG_CRIT, "Error parsing line #1 of expiry file %s", expiry_filename);
        while (fscanf(f, "%100s %u %u %u\n", table_name, &interval, &ops, &recs) == 4) {
            int in = expiry_size++;
            struct sc_expiry *expiry_new = realloc(expiry, expiry_size*sizeof(*expiry));
            if (expiry_new == NULL) {
                sc_log(LOG_CRIT, "Out of memory");
                expiry_size--;
                return;
            }
            expiry = expiry_new;
            strcpy(expiry[in].table_name, table_name);
            expiry[in].limit_ops = ops;
            expiry[in].next_ops = ops;
            expiry[in].limit_time = interval;
            expiry[in].last_time = time(NULL);
            expiry[in].next_time = expiry[in].last_time + interval;
            expiry[in].limit_records = recs;
            expiry[in].last_rowid = 1;
            line_no++;
        }
        if (expiry_size == 0 && schema_filename != NULL)
            sc_log(LOG_CRIT, "No expiry information could be parsed. I think there might be an error at line %u", line_no);
    }
}


/** Run the main (network protocol) loop of sqlcached */
static void start_server() {
    if (tcp_port != 0) {
        struct sockaddr_in sin;
        bzero(&sin, sizeof(sin));
        sock_tcp = socket(AF_INET, SOCK_STREAM, 0);
        if (sock_tcp < 0)
            sc_log(LOG_CRIT, "Cannot create TCP socket: %s", strerror(errno));
        sin.sin_family = AF_INET;
        sin.sin_addr.s_addr = htonl(INADDR_ANY);
        sin.sin_port = htons(tcp_port);
        if (bind(tcp_port, (struct sockaddr *)&sin, sizeof(sin)) < 0)
            sc_log(LOG_CRIT, "Cannot bind TCP port %d", tcp_port);
        if (listen(sock_tcp, 10) < 0)
            sc_log(LOG_CRIT, "Cannot listen() TCP socket: %s", strerror(errno));
        fcntl(sock_tcp, F_SETFL, O_NONBLOCK);
    }

    if (unix_socket_name != NULL) {
        struct sockaddr_un sun;
        bzero(&sun, sizeof(sun));
        sock_unix = socket(AF_LOCAL, SOCK_STREAM, 0);
        if (sock_unix < 0)
            sc_log(LOG_CRIT, "Cannot create Unix socket: %s", strerror(errno));
        strcpy(sun.sun_path, unix_socket_name);
        sun.sun_family = AF_LOCAL;
#ifdef __BSD_VISIBLE
        sun.sun_len = strlen(sun.sun_path);
#endif
        if (bind(sock_unix, (struct sockaddr *)&sun, SUN_LEN(&sun)) < 0)
            sc_log(LOG_CRIT, "Cannot bind Unix socket to %s: %s", sun.sun_path, strerror(errno));
        if (listen(sock_unix, 10) < 0)
            sc_log(LOG_CRIT, "Cannot listen() Unix socket: %s", strerror(errno));
        fcntl(sock_unix, F_SETFL, O_NONBLOCK);
    }

    conn = malloc(2*sizeof(*conn));
    if (conn == NULL)
        sc_log(LOG_CRIT, "Cannot allocate server structs: %s", strerror(errno));
    bzero(conn, 2*sizeof(*conn));

    conn[0].type = SC_CONN_TYPE_SERVER_TCP;
    conn[0].socket = sock_tcp;
    conn[1].type = SC_CONN_TYPE_SERVER_UNIX;
    conn[1].socket = sock_unix;
    conn_size = 2;
}


/** Stops the server */
static void stop_server() {
    if (sock_unix >= 0) {
        close(sock_unix);
        unlink(unix_socket_name);   /* needed??? */
    }
    if (sock_tcp >= 0)
        close(sock_tcp);
}


/** Stops the database */
static void stop_database() {
    sqlite_close(sqlite_db);
}


/* Forks a child to background */
static void daemonize() {
    int fd;
    
    /* Daemonize */
    switch (fork()) {
    case -1:
        fprintf(stderr, "Error forking to background\n");
        exit(1);
        break;
    case 0:
        break;
    default:
        exit(0);
    }
    
    if ((fd = open(_PATH_DEVNULL, O_RDWR, 0)) != -1) {
        /* Redirect stdin, stdout & stderr to /dev/null */
        dup2(fd, STDIN_FILENO);
        dup2(fd, STDOUT_FILENO);
        dup2(fd, STDERR_FILENO);
        if (fd > 2)
            close(fd);
    }
}

/*
static void dump_conn() {
    int i;
    printf("\n");
    for (i = 0; i < num_clients+2; i++)
        printf("%2d: %d %d\n", i, conn[i].type, conn[i].socket);
}
*/


/** Return index of a free conn array slot (enlarges the array 
 * if needed) */
static unsigned int sc_new_conn(int type, int socket) {
    int i;

    /* Try to find an empty slot */
    for (i = 0; i < conn_size; i++)
        if (conn[i].type == SC_CONN_TYPE_NO) {
            bzero(&conn[i], sizeof(*conn));
            conn[i].type = type;
            conn[i].socket = socket;
            conn[i].pstate = E_SC_START;
            return i;
        }

    conn_size++;
    conn = realloc(conn, conn_size*sizeof(*conn));
    bzero(&conn[conn_size-1], sizeof(*conn));
    conn[conn_size-1].type = type;
    conn[conn_size-1].socket = socket;
    conn[conn_size-1].pstate = E_SC_START;

    return conn_size-1;
}


static void sc_close_conn(int i) {
    close(conn[i].socket);
    sc_destroy_conn(&conn[i]);
    conn[i].socket = -1;
    conn[i].type = SC_CONN_TYPE_NO;
}


/** Handle SIGINT */
static void sigint_handler(int sig) {
    stop_server();
    stop_database();
    switch(sig) {
        case SIGINT:
            sc_log(LOG_ALERT, "Interrupted");
            break;
        case SIGHUP:
            sc_log(LOG_ALERT, "Hanged up");
            break;
        case SIGTERM:
            sc_log(LOG_ALERT, "Terminated");
            break;
        case SIGKILL:
            sc_log(LOG_ALERT, "Killed");
            break;
        default:
            sc_log(LOG_ALERT, "Stopped by unknown signal");
    }
    exit(0);
}


/** Sets up signal handlers */
static void set_signal_handlers() {
    signal(SIGINT, sigint_handler);
    signal(SIGKILL, sigint_handler);
    signal(SIGTERM, sigint_handler);
    signal(SIGHUP, sigint_handler);
    signal(SIGPIPE, SIG_IGN);
}


/** Runs sqlcached main loop / network comm. */
static void run_sqlcached() {
    int done = 0;
    struct pollfd *fds;
    int i, j;
    char *data_buf;
    ssize_t data_size;
    unsigned int old_conn_size;

    data_buf = malloc(READ_SIZE);
    if (data_buf == NULL)
        sc_log(LOG_CRIT, "Cannot allocate data_buf of %d", READ_SIZE);
   
    fds = malloc(conn_size*sizeof(*fds));
    if (fds == NULL)
        sc_log(LOG_CRIT, "Cannot allocate a small chunk of memory for poll fds. Go figure.");
    bzero(fds, conn_size*sizeof(*fds));

    old_conn_size = 0;
    while (!done) {
        if (conn_size != old_conn_size) {
            fds = realloc(fds, conn_size*sizeof(*fds));
            if (fds == NULL)
                sc_log(LOG_CRIT, "Cannot reallocate polls fds (to %d)", conn_size*sizeof(*fds));
            old_conn_size = conn_size;
        }
        for (i = 0; i < conn_size; i++) {
            fds[i].fd = conn[i].socket;
            fds[i].events = POLLIN;
            if (conn[i].outbuf_len > 0)
                fds[i].events |= POLLOUT;
            fds[i].revents = 0;
        }
        j = poll(fds, conn_size, 10000);
        if (j == -1)
            sc_log(LOG_CRIT, "Critical error in poll(): %s", strerror(errno));
        else if (j == 0)
            continue;
        /*
        if (debug_mode) {
            printf("poll %d: ", j);
            for (i = 0; i < conn_size; i++)
                printf(" %d", fds[i].revents);
            printf("\n");
        }
        */
        for (i = 2; i < conn_size; i++) { /* Do regular connections */
            if (conn[i].socket != -1 && (fds[i].revents & POLLIN) != 0) { /* Data to receive */
                data_size = recv(conn[i].socket, data_buf, READ_SIZE, /*MSG_DONTWAIT*/0);
                if (data_size <= 0) {
                    /* Assume the connection is (being) closed */
                    sc_close_conn(i);
                    sc_log(LOG_NOTICE, "Closed connection %p (%d)", &conn[i], i);
                    continue;
                }
                /*for (j = 0; j < data_size; j++)
                    sc_add_char(&conn[i], data_buf[j]);*/
                sc_add_char(&conn[i], data_buf, data_size);
            }
            if (conn[i].socket != -1 && (fds[i].revents & POLLOUT) != 0 && conn[i].outbuf_len > 0) { /* Data to send */
                int sent_size = send(conn[i].socket, conn[i].outbuf, conn[i].outbuf_len, 0);
                if (sent_size < 0) {
                    sc_log(LOG_ERR, "Error sending data on %p: %s", &conn[i], strerror(errno));
                    sc_close_conn(i);
                    continue;
                }
                /*fprintf(stderr, "Sent %d of %d\n", sent_size, conn[i].outbuf_len);*/
                if (sent_size != conn[i].outbuf_len)
                    bcopy(&conn[i].outbuf[sent_size], conn[i].outbuf, conn[i].outbuf_len - sent_size);
                conn[i].outbuf_len -= sent_size;
                if (conn[i].outbuf_len == 0 && (conn[i].flags & SC_CLOSE_ON_DONE_WRITING) != 0) {
                    sc_close_conn(i);
                    continue;
                }
            }
            if (conn[i].socket != -1 && (fds[i].revents & (POLLERR | POLLHUP | POLLNVAL)) != 0)  {
                sc_log(LOG_ERR, "Unhandled event %x on %p; closing connection.", fds[i].revents, &conn[i]);
                if ((fds[i].revents & POLLERR) != 0)
                    sc_log(LOG_ERR, "^^ POLLERR");
                if ((fds[i].revents & POLLHUP) != 0)
                    sc_log(LOG_ERR, "^^ POLLHUP");
                if ((fds[i].revents & POLLNVAL) != 0)
                    sc_log(LOG_ERR, "^^ POLLNVAL");
                sc_close_conn(i);
            }
        }
        if (conn[0].socket != -1 && (fds[0].revents & POLLIN) != 0) { /* tcp accept */
            struct sockaddr_in sin;
            unsigned int sin_len = sizeof sin;
            int sock_client;
            if ((sock_client = accept(fds[0].fd, (struct sockaddr*)&sin, &sin_len)) < 0)
                sc_log(LOG_CRIT, "Error accepting TCP client: %s", strerror(errno));
            else {
                int new_conn = sc_new_conn(SC_CONN_TYPE_CLIENT_TCP, sock_client);
                num_clients++;
                bcopy(&sin, &conn[new_conn].sin, sizeof sin);
                sc_log(LOG_NOTICE, "Accepted TCP client %p (%d)", &conn[new_conn], new_conn);
            }
        }
        if (conn[1].socket != -1 && (fds[1].revents & POLLIN) != 0) { /* unix accept */
            struct sockaddr_un sun;
            unsigned int sun_len = sizeof sun;
            int sock_client;
            if ((sock_client = accept(fds[1].fd, (struct sockaddr*)&sun, &sun_len)) < 0)
                sc_log(LOG_ERR, "Error accepting Unix client: %s", strerror(errno));
            else {
                int new_conn = sc_new_conn(SC_CONN_TYPE_CLIENT_UNIX, sock_client);
                num_clients++;
                bcopy(&sun, &conn[new_conn].sun, sizeof sun);
                sc_log(LOG_NOTICE, "Accepted Unix client %p (%d)", &conn[new_conn], new_conn);
            }
        }
    }
}


/** sqlcached main function */
int main(int _argc, char** _argv) {
    argc = _argc;
    argv = _argv;
    parse_args();
    
    if (!debug_mode)
        daemonize();

    umask(0111);
    set_signal_handlers();
    sc_log_init();
    start_database();
    start_server();
    run_sqlcached();
    stop_server();
    stop_database();
    
    return 0;
}
