/*-
 * Copyright (c) 2006. Ivan Voras <ivoras@gmail.com>
 * Released under the Artistic License
 */

/*
 * sqlcached server network protocol handler / database access
 * This is where the action is.
 * $Id: server_protocol.c,v 1.10 2006/11/23 11:37:56 ivoras Exp $
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>

#ifdef __linux__
/* Linux sucks. */
#define strlcpy(dst, src, size) strcpy(dst, src)
/* Theortically, defining _GNU_SOURCE should bring in definitions
 * of vasprintf and relatives, but it doesn't work. Typical.
 * For now, just ignore the compiler warnings.
 */
#endif
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <stdarg.h>
#include <time.h>

#include <sqlite.h>

#define SQLCACHED_SERVER
#include "common.h"

/* Originally BUF_SIZE=256, but that would overflow the 256-byte bucket */
#define BUF_SIZE 200

void sc_log(int prio, char *msg, ...);
void sc_do_command(struct sc_conn *sconn);
void sc_expire_data(int i, int why);

extern int debug_mode;
extern sqlite *sqlite_db;

extern struct sc_expiry *expiry;
extern unsigned int expiry_size;
extern unsigned int op_count;
extern unsigned int expiry_interval;

struct sc_call_state {
    struct sc_conn *sconn;
    int sent_cols;
    unsigned int n_cols;
    unsigned int n_rows;
};

enum SC_EXPIRE_WHY { EXPIRE_TIME, EXPIRE_OPS, EXPIRE_RECORDS };


/** Simple state machine that parses protocol command line received from
 * a client
 */
void sc_add_char(struct sc_conn *sconn, char *buf, ssize_t data_size) {
    char ch;
    ssize_t i;

    for (i = 0; i < data_size; i++) {
    ch = *(buf++);
start:
    switch (sconn->pstate) {
    case E_SC_START:
        if (isalnum(ch)) {
            sconn->pstate = E_SC_CMD_CHAR;
            sconn->cmd_pos = 0;
            goto start;
        } else
            sconn->pstate = E_SC_EAT_EOL;
        break;
    case E_SC_CMD_CHAR:
        sconn->cmd[sconn->cmd_pos++] = toupper(ch);
        if (sconn->cmd_pos == 3)
            sconn->pstate = E_SC_CMD_SPACE;
        break;
    case E_SC_EAT_EOL:
        if (ch == '\n' || ch == '\r')
            sconn->pstate = E_SC_EOL;
        break;
    case E_SC_EOL:
        if (ch != '\r' && ch != '\n') {
            sconn->pstate = E_SC_START;
            goto start;
        }
        break;
    case E_SC_CMD_SPACE:
        if (!isspace(ch)) {
            sconn->pstate = E_SC_PARAM_CHAR;
            goto start;
        }
        break;
    case E_SC_PARAM_CHAR:
        if (ch == '\r' || ch == '\n') {
            //sconn->param[sconn->param_len] = '\0';
            sc_do_command(sconn);
            sconn->pstate = E_SC_EOL;
            sconn->param_len = 0;
            sconn->param[0] = '\0';
            goto start;
        }
        if (ch == '\\' && (sconn->flags & SC_ESCAPE) == 0) {
            sconn->flags |= SC_ESCAPE;
            break;
        }
        sconn->param_len++;
        if (sconn->param_len+1 > sconn->param_size) {
            //unsigned int old_size = sconn->param_size;
            sconn->param_size += BUF_SIZE;
            sconn->param = realloc(sconn->param, sconn->param_size);
            if (sconn->param == NULL)
                sc_log(LOG_CRIT, "Cannot allocate param buffer for connection %p (%d bytes)", sconn, sconn->param_size);
            //bzero(sconn->param + old_size, BUF_SIZE);
        }
        if ((sconn->flags & SC_ESCAPE) != 0) {
            switch (ch) {
                case 'n':
                    ch = '\n';
                    break;
                case 'r':
                    ch = '\r';
                    break;
                case 't':
                    ch = '\t';
                    break;
                case '0':
                    ch = '\0';
                    break;
                case '\\':
                    ch = '\\';
                    break;
                case '\'':
                    ch = '\'';
                    break;
                case 'a':
                    ch = '\a';
                    break;
                case 'b':
                    ch = '\b';
                    break;
            }
            sconn->flags &= ~SC_ESCAPE;
        }
        sconn->param[sconn->param_len-1] = ch;
        sconn->param[sconn->param_len] = '\0';
        break;
    }
    /*printf("%p %c\n", sconn, ch);*/
    }
}


void sc_destroy_conn(struct sc_conn *sconn) {
    if (sconn->param != NULL) {
        free(sconn->param);
        sconn->param = NULL;
        sconn->param_len = sconn->param_size = 0;
    }
    if (sconn->outbuf != NULL) {
        free(sconn->outbuf);
        sconn->outbuf = NULL;
        sconn->outbuf_len = sconn->outbuf_size = 0;
    }
}


void sc_add_outbuf(struct sc_conn *sconn, char *buf, unsigned int size) {
    if (sconn->outbuf == NULL) { /* Fresh buffer */
        sconn->outbuf = malloc(BUF_SIZE);
        if (sconn->outbuf == NULL)
            sc_log(LOG_CRIT, "Cannot allocate outbuf");
        sconn->outbuf_size = BUF_SIZE;
        sconn->outbuf_len = 0;
    } else if (sconn->outbuf_len + size > sconn->outbuf_size) { /* Extend the buffer */
        unsigned int new_size = sconn->outbuf_len + size;
        new_size = ((new_size / BUF_SIZE) + 1) * BUF_SIZE; /* Round it to high multiple of BUF_SIZE */
        sconn->outbuf_size = new_size;
        sconn->outbuf = realloc(sconn->outbuf, new_size);
        if (sconn->outbuf == NULL)
            sc_log(LOG_CRIT, "Cannot extend outbuf to %d", new_size);
    }
    bcopy(buf, &sconn->outbuf[sconn->outbuf_len], size);
    sconn->outbuf_len += size;
}


void sc_add_outf(struct sc_conn *sconn, char *msg, ...) {
    va_list ap;
    char *buf;
    int buf_size;

    va_start(ap, msg);
    buf_size = vasprintf(&buf, msg, ap);
    va_end(ap);

    if (buf_size <= 0)
        sc_log(LOG_CRIT, "Cannot create a formatted buffer at %s", msg);

    sc_add_outbuf(sconn, buf, buf_size);
    free(buf);
}


static int sc_db_callback(void *_sc, int n_cols, char **data, char **names) {
    struct sc_call_state *sc = (struct sc_call_state*)_sc;
    int i;
    static char *qbuf = NULL;
    static int qbuf_len = 100;

    if (!sc->sent_cols) {
        sc->n_cols = n_cols;
        for (i = 0; i < n_cols; i++)
            sc_add_outf(sc->sconn, "%s\r\n", names[i]);
        sc->sent_cols = 1;
    }
    sc->n_rows++;
   
    if (qbuf == NULL)
        qbuf = malloc(qbuf_len);
    
    for (i = 0; i < n_cols; i++) {
        char *c = qbuf;
        char *d = data[i];
        if (d == NULL) {
            sc_add_outf(sc->sconn, "\r\n");
            continue;
        }
        int ql = 0;
        do {
            /* Newline characters must be quoted */
            switch (*d) {
                case '\r':
                    *c++ = '\\';
                    *c++ = 'r';
                    ql++;
                    break;
                case '\n':
                    *c++ = '\\';
                    *c++ = 'n';
                    ql++;
                    break;
                default:
                    *c++ = *d;
            }
            ql++;
            if (ql >= qbuf_len - 2) {
                qbuf_len *= 2;
                qbuf = realloc(qbuf, qbuf_len);
                c = qbuf + ql;
            }
        } while (*d++ != '\0');
        sc_add_outf(sc->sconn, "%s\r\n", qbuf);
    }

    return 0;
}


void sc_do_command(struct sc_conn *sconn) {
    op_count++;
    
    if (strcmp(sconn->cmd, "SQL") == 0) {
        /* SQL query to execute */
        char *err;
        struct sc_call_state call_state;
        int n_rows_nc, n_cols_nc;

        bzero(&call_state, sizeof call_state);
        call_state.sconn = sconn;
        sc_add_outf(sconn, "+REC ");
        n_rows_nc = sconn->outbuf_len;
        sc_add_outf(sconn, "0000, ");
        n_cols_nc = sconn->outbuf_len;
        sc_add_outf(sconn, "0000\r\n");
        if (sqlite_exec(sqlite_db, sconn->param, sc_db_callback, &call_state, &err) == SQLITE_OK) {
            if (call_state.n_rows > 0) {
                char *n_rows_c, *n_cols_c;
                /* Reasonably ugly fixup, but efficient */
                n_rows_c = sconn->outbuf + n_rows_nc;
                n_cols_c = sconn->outbuf +  n_cols_nc;
                sprintf(n_rows_c, "%04d", call_state.n_rows);
                n_rows_c[4] = ',';
                sprintf(n_cols_c, "%04d", call_state.n_cols);
                n_cols_c[4] = '\r';
            } else {
                char *sql = sconn->param;
                sconn->outbuf_len = 0;
                while (sconn->param[0] == ' ' || sconn->param[0] == '\t') // skip whitespace
                    sql++;
                if (strncasecmp(sql, "SELECT", 6) == 0) 
                    sc_add_outf(sconn, "+NREC 0\r\n");
                else 
                    sc_add_outf(sconn, "+NREC %d\r\n", sqlite_changes(sqlite_db));
            }
        } else {
            char *m = err;
            while (*m != '\0') {
                if ( (*m == '\r') || (*m == '\n') )
                    *m = '#';
                m++;
                sc_log(LOG_ERR, err);
            }
            sconn->outbuf_len = 0;
            sc_add_outf(sconn, "-ERROR sqlite: %s\r\n", err);
        }
    } else if (strcmp(sconn->cmd, "VER") == 0) {
        /* Version handshake */
        if (strcmp(sconn->param, SC_VER_SIG) == 0) 
            sc_add_outf(sconn, "+VER %s\r\n", SC_VER_SIG);
        else {
            sc_add_outf(sconn, "-ERROR Invalid client version\r\n");
            sc_log(LOG_ERR, "Invalid handshake version: expecting %s, got %s", SC_VER_SIG, sconn->param);
            shutdown(sconn->socket, SHUT_RD);
            sconn->flags |= SC_CLOSE_ON_DONE_WRITING;
        }
    } else if (strcmp(sconn->cmd, "EXP") == 0) {
        /* Add expiry data */
        char table[101];
        int interval, ops, recs;
        if (sscanf(sconn->param, "%100s %u %u %u", table, &interval, &ops, &recs) == 4) {
            struct sc_expiry *sex = NULL;
            int i;
            for (i = 0; i < expiry_size; i++) {
                if (strncmp(table, expiry[i].table_name, 100) == 0) {
                    sex = &expiry[i];
                    break;
                }
            }
            if (sex == NULL) {
                /* An unfortunate state */
                i = expiry_size++;
                struct sc_expiry *expiry_new = realloc(expiry, expiry_size*sizeof(*expiry));
                if (expiry_new == NULL) {
                    sc_log(LOG_ERR, "Out of memory in EXP");
                    expiry_size--;
                    return;
                }
                expiry = expiry_new;
                sex = &expiry[i];
            }
            strlcpy(sex->table_name, table, 100);
            sex->limit_ops = ops;
            sex->next_ops = ops;
            sex->limit_time = interval;
            sex->last_time = time(NULL);
            sex->next_time = sex->last_time + interval;
            sex->limit_records = recs;
            sex->last_rowid = 1;
            sc_add_outf(sconn, "+OK expiry added as #%d\r\n", i);
        } else {
            sc_add_outf(sconn, "-ERROR Invalid expiry spec\r\n");
            sc_log(LOG_ERR, "Invalid expiry spec: %s", sconn->param);
        }
    } else if (strcmp(sconn->cmd, "EXI") == 0) {
        int ops;
        if (sscanf(sconn->param, "%d", &ops) == 1) {
            expiry_interval = ops;
            sc_add_outf(sconn, "+OK expiry ops interval set: %d\r\n", ops);
        } else {
            sc_add_outf(sconn, "-ERROR Invalid expiry spec\r\n");
            sc_log(LOG_ERR, "Invalid expiry ops interval spec: %s", sconn->param);
        }
    }

    sc_check_expiry();
}


void sc_check_expiry() {
    /* Check if something needs to be expired */
    if (expiry_interval != 0 && op_count % expiry_interval != 0)
        return;
    else {
        int i, did_expire = 0;
        unsigned int now = time(NULL);

        for (i = 0; i < expiry_size; i++) {
            if (expiry[i].next_ops <= op_count && expiry[i].limit_ops != 0) {
                /*printf("next_ops: %d, limit_ops: %d\n", expiry[i].next_ops, expiry[i].limit_ops);*/
                sc_expire_data(i, EXPIRE_OPS);
                did_expire = 1;
                expiry[i].next_ops += expiry[i].limit_ops;
                expiry[i].last_time = now;
            }
            if (expiry[i].next_time <= now && expiry[i].limit_time != 0) {
                if (!did_expire) {
                    sc_expire_data(i, EXPIRE_TIME);
                    did_expire = 1;
                }
                expiry[i].next_time = expiry[i].last_time + expiry[i].limit_time;
                expiry[i].last_time = now;
            }
            if (expiry[i].limit_records != 0)
                sc_expire_data(i, EXPIRE_RECORDS);
        }
    }
}


/** sqlite callback that fetches a single integer */
static int sc_db_int_callback(void *_i, int n_cols, char **data, char **names) {
    unsigned int *i = _i;
    if (n_cols != 1)
        sc_log(LOG_ERR, "sc_db_int_callback: n_cols=%d", n_cols);
    if (n_cols < 1)
        sc_log(LOG_CRIT, "sc_db_int_callback: no cols?");
    *i = atol(data[0]);
    return 0;
}


void sc_expire_data(int i, int why) {
    char *err;
    char sql[500];
    int res;

    if (why == EXPIRE_OPS || why == EXPIRE_TIME) {
        int nr;
        sprintf(sql, "DELETE FROM %s WHERE time >= %u", expiry[i].table_name, expiry[i].last_time);
        if (sqlite_exec(sqlite_db, sql, NULL, NULL, &err) != SQLITE_OK) {
            sc_log(LOG_ERR, "Cannot execute expiry sql: %s: %s", sql, err);
            free(err);
        }
        nr = sqlite_changes(sqlite_db);
        sc_log(LOG_NOTICE, "Expiring %d records by %s since %u", nr, why == EXPIRE_OPS ? "ops" : "time", expiry[i].last_time);
    } else if (why == EXPIRE_RECORDS) {
        unsigned int n_records;
        sprintf(sql, "SELECT count(*) FROM %s", expiry[i].table_name);
        res = sqlite_exec(sqlite_db, sql, sc_db_int_callback, &n_records, &err);
        if (res != SQLITE_OK) {
            sc_log(LOG_ERR, "Cannot fetch row count for expiry: %s: %s", sql, err);
            free(err);
        } else if (n_records > expiry[i].limit_records) {
            unsigned int max_rowid, min_rowid;
            signed long int cutoff_rowid;

            if (debug_mode) {
                sprintf(sql, "SELECT min(rowid) FROM %s", expiry[i].table_name);
                res = sqlite_exec(sqlite_db, sql, sc_db_int_callback, &min_rowid, &err);
                if (res != SQLITE_OK) {
                    sc_log(LOG_ERR, "Cannot fetch min rowid for expiry: %s: %s", sql, err);
                    free(err);
                    return;
                }
            }
            
            sprintf(sql, "SELECT max(rowid) FROM %s", expiry[i].table_name);
            res = sqlite_exec(sqlite_db, sql, sc_db_int_callback, &max_rowid, &err);
            if (res != SQLITE_OK) {
                sc_log(LOG_ERR, "Cannot fetch max rowid for expiry: %s: %s", sql, err);
                free(err);
                return;
            }
            if (max_rowid < expiry[i].limit_records + 2*expiry_interval)
                return;

            cutoff_rowid = max_rowid - expiry[i].limit_records - expiry_interval;
            if (cutoff_rowid < 0)
                cutoff_rowid = max_rowid - expiry[i].limit_records;
            sprintf(sql, "DELETE FROM %s WHERE rowid < %u", expiry[i].table_name, (unsigned int)cutoff_rowid);
            res = sqlite_exec(sqlite_db, sql, NULL, NULL, &err);
            if (res != SQLITE_OK) {
                sc_log(LOG_ERR, "Cannot execute expiry sql: %s: %s", sql, err);
                free(err);
                return;
            }

            if (debug_mode)
                sc_log(LOG_NOTICE, "Expiry by records: %u to %u", min_rowid, (unsigned int)cutoff_rowid);
            expiry[i].last_rowid = max_rowid;
        }
    }
}
